@@ -39,7 +39,7 @@ static const struct grub_arg_option options[] = {
   {0, 0, 0, 0, 0, 0}
 };
 
-static struct { const char *name; const char *hashname; } aliases[] = 
+static struct { const char *name; const char *hashname; } aliases[] =
   {
     {"sha256sum", "sha256"},
     {"sha512sum", "sha512"},
@@ -63,17 +63,23 @@ hextoval (char c)
 static grub_err_t
 hash_file (grub_file_t file, const gcry_md_spec_t *hash, void *result)
 {
-  GRUB_PROPERLY_ALIGNED_ARRAY (context, hash->contextsize);
-  grub_uint8_t readbuf[4096];
+  void *context;
+  grub_uint8_t *readbuf;
+#define BUF_SIZE 4096
+  readbuf = grub_malloc (BUF_SIZE);
+  if (!readbuf)
+    return grub_errno;
+  context = grub_zalloc (hash->contextsize);
+  if (!readbuf || !context)
+    goto fail;
 
-  grub_memset (context, 0, sizeof (context));
-  hash->init (context);
+  hash->init (context, 0);
   while (1)
     {
       grub_ssize_t r;
-      r = grub_file_read (file, readbuf, sizeof (readbuf));
+      r = grub_file_read (file, readbuf, BUF_SIZE);
       if (r < 0)
-	return grub_errno;
+	goto fail;
       if (r == 0)
 	break;
       hash->write (context, readbuf, r);
@@ -81,7 +87,15 @@ hash_file (grub_file_t file, const gcry_md_spec_t *hash, void *result)
   hash->final (context);
   grub_memcpy (result, hash->read (context), hash->mdlen);
 
+  grub_free (readbuf);
+  grub_free (context);
+
   return GRUB_ERR_NONE;
+
+ fail:
+  grub_free (readbuf);
+  grub_free (context);
+  return grub_errno;
 }
 
 static grub_err_t
@@ -90,16 +104,19 @@ check_list (const gcry_md_spec_t *hash, const char *hashfilename,
 {
   grub_file_t hashlist, file;
   char *buf = NULL;
-  grub_uint8_t expected[hash->mdlen];
-  grub_uint8_t actual[hash->mdlen];
+  grub_uint8_t expected[GRUB_CRYPTO_MAX_MDLEN];
+  grub_uint8_t actual[GRUB_CRYPTO_MAX_MDLEN];
   grub_err_t err;
   unsigned i;
   unsigned unread = 0, mismatch = 0;
 
-  hashlist = grub_file_open (hashfilename);
+  if (hash->mdlen > GRUB_CRYPTO_MAX_MDLEN)
+    return grub_error (GRUB_ERR_BUG, "mdlen is too long");
+
+  hashlist = grub_file_open (hashfilename, GRUB_FILE_TYPE_HASHLIST);
   if (!hashlist)
     return grub_errno;
-  
+
   while (grub_free (buf), (buf = grub_file_getline (hashlist)))
     {
       const char *p = buf;
@@ -111,30 +128,37 @@ check_list (const gcry_md_spec_t *hash, const char *hashfilename,
 	  high = hextoval (*p++);
 	  low = hextoval (*p++);
 	  if (high < 0 || low < 0)
-	    return grub_error (GRUB_ERR_BAD_FILE_TYPE, "invalid hash list");
+	    {
+	      grub_free (buf);
+	      return grub_error (GRUB_ERR_BAD_FILE_TYPE, "invalid hash list");
+	    }
 	  expected[i] = (high << 4) | low;
 	}
       if ((p[0] != ' ' && p[0] != '\t') || (p[1] != ' ' && p[1] != '\t'))
-	return grub_error (GRUB_ERR_BAD_FILE_TYPE, "invalid hash list");
+	{
+	  grub_free (buf);
+	  return grub_error (GRUB_ERR_BAD_FILE_TYPE, "invalid hash list");
+	}
       p += 2;
       if (prefix)
 	{
 	  char *filename;
-	  
+
 	  filename = grub_xasprintf ("%s/%s", prefix, p);
 	  if (!filename)
-	    return grub_errno;
-	  if (!uncompress)
-	    grub_file_filter_disable_compression ();
-	  file = grub_file_open (filename);
+	    {
+	      grub_free (buf);
+	      return grub_errno;
+	    }
+	  file = grub_file_open (filename, GRUB_FILE_TYPE_TO_HASH
+				 | (!uncompress ? GRUB_FILE_TYPE_NO_DECOMPRESS
+				    : GRUB_FILE_TYPE_NONE));
 	  grub_free (filename);
 	}
       else
-	{
-	  if (!uncompress)
-	    grub_file_filter_disable_compression ();
-	  file = grub_file_open (p);
-	}
+	file = grub_file_open (p, GRUB_FILE_TYPE_TO_HASH
+			       | (!uncompress ? GRUB_FILE_TYPE_NO_DECOMPRESS
+				  : GRUB_FILE_TYPE_NONE));
       if (!file)
 	{
 	  grub_file_close (hashlist);
@@ -168,7 +192,7 @@ check_list (const gcry_md_spec_t *hash, const char *hashfilename,
 				 "hash of '%s' mismatches", p);
 	    }
 	  mismatch++;
-	  continue;	  
+	  continue;
 	}
       grub_printf_ (N_("%s: OK\n"), p);
     }
@@ -205,6 +229,9 @@ grub_cmd_hashsum (struct grub_extcmd_context *ctxt,
   if (!hash)
     return grub_error (GRUB_ERR_BAD_ARGUMENT, "unknown hash");
 
+  if (hash->mdlen > GRUB_CRYPTO_MAX_MDLEN)
+    return grub_error (GRUB_ERR_BUG, "mdlen is too long");
+
   if (state[2].set)
     prefix = state[2].arg;
 
@@ -218,13 +245,13 @@ grub_cmd_hashsum (struct grub_extcmd_context *ctxt,
 
   for (i = 0; i < (unsigned) argc; i++)
     {
-      GRUB_PROPERLY_ALIGNED_ARRAY (result, hash->mdlen);
+      GRUB_PROPERLY_ALIGNED_ARRAY (result, GRUB_CRYPTO_MAX_MDLEN);
       grub_file_t file;
       grub_err_t err;
       unsigned j;
-      if (!uncompress)
-	grub_file_filter_disable_compression ();
-      file = grub_file_open (args[i]);
+      file = grub_file_open (args[i], GRUB_FILE_TYPE_TO_HASH
+			     | (!uncompress ? GRUB_FILE_TYPE_NO_DECOMPRESS
+				: GRUB_FILE_TYPE_NONE));
       if (!file)
 	{
 	  if (!keep)
