@@ -419,7 +419,7 @@ read_expression (struct parsebuf *p)
   skip_whitespace (p);
   if (peek_char (p) == '"')
     {
-      /* Read as a quoted string.  
+      /* Read as a quoted string.
          The quotation marks are not included in the expression value.  */
       /* Skip opening quotation mark.  */
       read_char (p);
@@ -484,7 +484,7 @@ parse_proportional_spec (const char *value, signed *abs, grub_fixed_signed_t *pr
 	  ptr++;
 	}
 
-      num = grub_strtoul (ptr, (char **) &ptr, 0);
+      num = grub_strtoul (ptr, &ptr, 0);
       if (grub_errno)
 	return grub_errno;
       if (sig)
