@@ -168,36 +168,27 @@ iterate_dev (const char *devname, void *data __attribute__ ((unused)))
       grub_errno = GRUB_ERR_NONE;
       return 0;
     }
-  {
-    char tmp[grub_strlen (devname) + sizeof (",")];
 
-    grub_memcpy (tmp, devname, grub_strlen (devname));
+  if (grub_strcmp (devname, current_word) == 0)
+    {
+      if (add_completion (devname, ")", GRUB_COMPLETION_TYPE_PARTITION))
+	{
+	  grub_device_close (dev);
+	  return 1;
+	}
 
-    if (grub_strcmp (devname, current_word) == 0)
-      {
-	if (add_completion (devname, ")", GRUB_COMPLETION_TYPE_PARTITION))
+      if (dev->disk)
+	if (grub_partition_iterate (dev->disk, iterate_partition, NULL))
 	  {
 	    grub_device_close (dev);
 	    return 1;
 	  }
-
-	if (dev->disk)
-	  if (grub_partition_iterate (dev->disk, iterate_partition, NULL))
-	    {
-	      grub_device_close (dev);
-	      return 1;
-	    }
-      }
-    else
-      {
-	grub_memcpy (tmp + grub_strlen (devname), "", sizeof (""));
-	if (add_completion (tmp, "", GRUB_COMPLETION_TYPE_DEVICE))
-	  {
-	    grub_device_close (dev);
-	    return 1;
-	  }
-      }
-  }
+    }
+  else if (add_completion (devname, "", GRUB_COMPLETION_TYPE_DEVICE))
+    {
+      grub_device_close (dev);
+      return 1;
+    }
 
   grub_device_close (dev);
   grub_errno = GRUB_ERR_NONE;
@@ -293,10 +284,11 @@ complete_file (void)
 
       /* Cut away the filename part.  */
       dirfile = grub_strrchr (dir, '/');
-      dirfile[1] = '\0';
+      if (dirfile)
+	dirfile[1] = '\0';
 
       /* Iterate the directory.  */
-      (fs->dir) (dev, dir, iterate_dir, NULL);
+      (fs->fs_dir) (dev, dir, iterate_dir, NULL);
 
       grub_free (dir);
 
@@ -409,8 +401,8 @@ char *
 grub_normal_do_completion (char *buf, int *restore,
 			   void (*hook) (const char *, grub_completion_type_t, int))
 {
-  int argc;
-  char **argv;
+  int argc = 0;
+  char **argv = NULL;
 
   /* Initialize variables.  */
   match = 0;
@@ -498,6 +490,9 @@ grub_normal_do_completion (char *buf, int *restore,
 	  spaces++;
 
       ret = grub_malloc (match_len - current_len + grub_strlen (suffix) + spaces + 1);
+      if (ret == NULL)
+        goto fail;
+
       newstr = ret;
       for (escstr = match + current_len; *escstr; escstr++)
 	{
@@ -509,7 +504,7 @@ grub_normal_do_completion (char *buf, int *restore,
       *newstr = '\0';
 
       if (num_found == 1)
-	grub_strcat (ret, suffix);
+	grub_strcpy (newstr, suffix);
 
       if (*ret == '\0')
 	{
@@ -525,10 +520,8 @@ grub_normal_do_completion (char *buf, int *restore,
 
  fail:
   if (argc != 0)
-    {
-      grub_free (argv[0]);
-      grub_free (argv);
-    }
+    grub_free (argv[0]);
+  grub_free (argv);
   grub_free (match);
   grub_errno = GRUB_ERR_NONE;
 
