@@ -21,26 +21,12 @@
 
 #include <grub/macho.h>
 
-#define GRUB_MACHO_CPUTYPE_IS_HOST32(x) ((x)==0x00000007)
-#define GRUB_MACHO_CPUTYPE_IS_HOST64(x) ((x)==0x01000007)
-
-struct grub_macho_thread32
-{
-  grub_uint32_t cmd;
-  grub_uint32_t cmdsize;
-  grub_uint8_t unknown1[48];
-  grub_uint32_t entry_point;
-  grub_uint8_t unknown2[20];
-} __attribute__ ((packed));
-
-
-struct grub_macho_thread64
-{
-  grub_uint32_t cmd;
-  grub_uint32_t cmdsize;
-  grub_uint8_t unknown1[0x88];
-  grub_uint64_t entry_point;
-  grub_uint8_t unknown2[0x20];
-} __attribute__ ((packed));
+#define GRUB_MACHO_CPUTYPE_IS_HOST32(x) ((x) == GRUB_MACHO_CPUTYPE_IA32)
+#define GRUB_MACHO_CPUTYPE_IS_HOST64(x) ((x) == GRUB_MACHO_CPUTYPE_AMD64)
+#ifdef __x86_64__
+#define GRUB_MACHO_CPUTYPE_IS_HOST_CURRENT(x) ((x) == GRUB_MACHO_CPUTYPE_AMD64)
+#else
+#define GRUB_MACHO_CPUTYPE_IS_HOST_CURRENT(x) ((x) == GRUB_MACHO_CPUTYPE_IA32)
+#endif
 
 #endif
